from math import floor, sqrt
from typing import List, Tuple

import numpy as np
from scipy.ndimage import gaussian_filter
from scipy.stats import entropy, multivariate_normal

from . import ILaneFilter

from .types import Segment, SegmentColor


class LaneFilterHistogram(ILaneFilter):
    """Generates an estimate of the lane pose.

    Creates and maintain a histogram grid filter to estimate the lane pose.

    Lane pose is defined as the tuple (`d`, `phi`) : lateral deviation and angulare deviation
    from the center of the lane.

    - Predict step : Uses the estimated linear and angular velocities to predict the change in
    the lane pose.

    - Update Step : The filter receives a segment list. For each segment, it extracts the
    corresponding lane pose "votes", and adds it to the corresponding part of the histogram.

    Best estimate correspond to the slot of the histogram with the highest voted value.
    """

    def __init__(self, *args, **kwargs):
        super(LaneFilterHistogram, self).__init__(*args, **kwargs)

        # Additional variables
        self.red_to_white = False
        self.use_yellow = True
        self.range_est_min = 0
        self.filtered_segments = []
        # initialize
        self.initialize()

    def initialize(self):
        pos = np.empty(self.d.shape + (2,))
        pos[:, :, 0] = self.d
        pos[:, :, 1] = self.phi
        RV = multivariate_normal(self.mean_0, self.cov_0)
        self.belief = RV.pdf(pos)

    def get_entropy(self) -> float:
        belief = self.belief
        s = entropy(belief.flatten())
        return s

    def predict(self,left_encoder_ticks: float, right_encoder_ticks: float):
        # Calculate v and w from ticks using kinematics
        R = self.wheel_radius
        alpha = 2 * np.pi / self.encoder_resolution
        d_left = R * alpha * left_encoder_ticks
        d_right = R * alpha * right_encoder_ticks
        d_A = (d_left + d_right) / 2
        w = (d_right - d_left) / self.wheel_baseline
        [d_t, phi_t] = self.get_estimate()
        v = d_A * np.sin(w + phi_t)

        # Propagate each centroid forward using the kinematic function
        d_t = self.d + v
        phi_t = self.phi + w

        p_belief = np.zeros(self.belief.shape)

        # there has got to be a better/cleaner way to do this - just applying the process model to
        # translate each cell value
        for i in range(self.belief.shape[0]):
            for j in range(self.belief.shape[1]):
                if self.belief[i, j] > 0:
                    if (
                        d_t[i, j] > self.d_max
                        or d_t[i, j] < self.d_min
                        or phi_t[i, j] < self.phi_min
                        or phi_t[i, j] > self.phi_max
                    ):
                        continue

                    i_new = int(floor((d_t[i, j] - self.d_min) / self.delta_d))
                    j_new = int(floor((phi_t[i, j] - self.phi_min) / self.delta_phi))

                    p_belief[i_new, j_new] += self.belief[i, j]

        s_belief = np.zeros(self.belief.shape)
        gaussian_filter(p_belief, self.cov_mask, output=s_belief, mode="constant")

        if np.sum(s_belief) == 0:
            return

        self.belief = s_belief / np.sum(s_belief)

    def update(self, segments: List[Segment]):
        # prepare the segments for each belief array
        segmentsArray = self._prepare_segments(segments)
        # generate all belief arrays

        measurement_likelihood = self._generate_measurement_likelihood(segmentsArray)
        if measurement_likelihood is not None:
            self.belief = np.multiply(self.belief, measurement_likelihood)
            if np.sum(self.belief) == 0:
                self.belief = measurement_likelihood
            else:
                self.belief /= np.sum(self.belief)

    
    def get_estimate(self) -> Tuple[float, float]:
        maxids = np.unravel_index(self.belief.argmax(), self.belief.shape)
        d_max = self.d_min + (maxids[0] + 0.5) * self.delta_d
        phi_max = self.phi_min + (maxids[1] + 0.5) * self.delta_phi

        return d_max, phi_max


    def get_max(self) -> float:
        return self.belief.max()

    def get_inlier_segments(self, segments: List[Segment], d_max, phi_max):
        inlier_segments = []
        for segment in segments:
            d_s, phi_s, l, w = self._generate_vote(segment)
            if abs(d_s - d_max) < self.delta_d and abs(phi_s - phi_max) < self.delta_phi:
                inlier_segments.append(segment)
        return inlier_segments

    def _generate_measurement_likelihood(self, segments: List[Segment]):
        # initialize measurement likelihood to all zeros
        measurement_likelihood = np.zeros(self.d.shape)

        for segment in segments:
            d_i, phi_i = self._generate_vote(segment)

            # if the vote lands outside of the histogram discard it
            if (d_i > self.d_max
                    or d_i < self.d_min
                    or phi_i < self.phi_min
                    or phi_i > self.phi_max
            ):
                continue

            i = int(floor((d_i - self.d_min) / self.delta_d))
            j = int(floor((phi_i - self.phi_min) / self.delta_phi))
            measurement_likelihood[i, j] += 1

        if np.linalg.norm(measurement_likelihood) == 0:
            return None
        measurement_likelihood /= np.sum(measurement_likelihood)
        return measurement_likelihood

    # generate a vote for one segment
    def _generate_vote(self, segment: Segment):
        p1 = segment.points[0].as_array()
        p2 = segment.points[1].as_array()
        t_hat = (p2 - p1) / np.linalg.norm(p2 - p1)

        n_hat = np.array([-t_hat[1], t_hat[0]])
        d1 = np.inner(n_hat, p1)
        d2 = np.inner(n_hat, p2)
        l1 = np.inner(t_hat, p1)
        l2 = np.inner(t_hat, p2)
        if l1 < 0:
            l1 = -l1
        if l2 < 0:
            l2 = -l2

        l_i = (l1 + l2) / 2
        d_i = (d1 + d2) / 2
        phi_i = np.arcsin(t_hat[1])
        if segment.color == SegmentColor.WHITE:  # right lane is white
            if p1[0] > p2[0]:  # right edge of white lane
                d_i -= self.linewidth_white
            else:  # left edge of white lane
                d_i = -d_i
                phi_i = -phi_i
            d_i -= self.lanewidth / 2

        elif segment.color == SegmentColor.YELLOW:  # left lane is yellow
            if p2[0] > p1[0]:  # left edge of yellow lane
                d_i -= self.linewidth_yellow
                phi_i = -phi_i
            else:  # right edge of white lane
                d_i = -d_i

            d_i = self.lanewidth/2 - d_i

        return d_i, phi_i

    @staticmethod
    def _get_segment_distance(segment: Segment):
        """
        Gets the distance from the center of the robot to the center point of the given segment.

        Args:
            segment:

        Returns:

        """
        x_c = (segment.points[0].x + segment.points[1].x) / 2
        y_c = (segment.points[0].y + segment.points[1].y) / 2
        return sqrt(x_c ** 2 + y_c ** 2)

    def _prepare_segments(self, segments: List[Segment]):
        """
        Prepares the segments for the creation of the belief arrays.

        Args:
            segments:

        Returns:

        """
        filtered_segments = []
        for segment in segments:
            # Optional transform from RED to WHITE
            if self.red_to_white and segment.color == SegmentColor.RED:
                segment.color = SegmentColor.WHITE

            # Optional filtering out YELLOW
            if not self.use_yellow and segment.color == SegmentColor.YELLOW:
                continue

            # we don't care about RED ones for now
            if segment.color != SegmentColor.WHITE and segment.color != SegmentColor.YELLOW:
                continue
            # filter out any segments that are behind us
            if segment.points[0].x < 0 or segment.points[1].x < 0:
                continue

            # only consider points in a certain range from the robot for the position estimation
            # point_range = self._get_segment_distance(segment)
            # if self.range_est > point_range > self.range_est_min:
            filtered_segments.append(segment)

        return filtered_segments


__all__ = ["LaneFilterHistogram"]



import io
import os

import numpy as np
import numpy.ma as ma
from reprep.graphics.filter_scale import scale

import cv2
import matplotlib.pyplot as plt

from dt_state_estimation.lane_filter import ILaneFilter
from typing import Tuple


BGRImage = np.ndarray
assets_dir: str = os.path.abspath(os.path.join(os.path.dirname(__file__), "assets"))
background_image_fpath: str = os.path.join(assets_dir, "lane_filter_grid.png")
background_image: BGRImage = cv2.imread(background_image_fpath)


def plot_d_phi(d: float, phi: float, size: Tuple[int, int] = (-1, -1)):
    """
    Generates a debug image with the estimated pose of the robot drawn.

    Args:
        d (:obj:`float`): Estimated `d`
        phi (:obj:`float`): Estimated `phi`
        size (:obj:`tuple`): Size of the image to draw

    Returns:
        :obj:`numpy array`: an OpenCV image

    """
    size_x, size_y = size
    # 36cm is the total width depicted in the background image
    total_width_meters = 0.36
    # in the image, the origin is at 23.7cm from the left border
    origin_meters = 0.237

    # the axis of the rendered image and the axis of the robot are flipped
    d *= -1.0
    phi *= -1.0

    # start with a fresh copy of the background
    image = background_image.copy()

    # resize if needed
    if size_x > 0 and size_y > 0:
        image = cv2.resize(image, size)
    size_y, size_x, _ = image.shape

    # compute origin
    pixel_per_meter = size_x / total_width_meters
    origin_x_px = origin_meters * pixel_per_meter
    origin_y_px = size_y * 0.5

    # compute robot's location
    x = int(origin_x_px + d * pixel_per_meter)
    y = int(origin_y_px)

    # draw location
    cv2.circle(image, (x, y), 14, (0, 0, 200), 3)

    # simple 2D rotation, used to rotate the heading indicator
    R = np.array([
        [np.cos(phi), -np.sin(phi), 0],
        [np.sin(phi), np.cos(phi), 0],
        [0, 0, 1]
    ])

    # compute heading indicator line coordinates
    heading_bar = np.array([[0, 0, 1], [0, -40, 1]])
    heading_bar = np.dot(R, heading_bar.T).T[:, :2].astype(int) + [x, y]

    # draw heading
    cv2.line(image, tuple(heading_bar[0]), tuple(heading_bar[1]), (0, 0, 200), 5)
    # ---
    return image


def plot_belief(
        filter: ILaneFilter,
        bgcolor=(0, 204, 255),
        dpi=150,
        other_phi=None,
        other_d=None,
) -> BGRImage:
    """Returns a BGR image"""

    # get estimate
    d, phi = filter.get_estimate()

    bgcolor = tuple(x / 255.0 for x in bgcolor)
    figure = plt.figure(facecolor=bgcolor)

    f_d = lambda x: 100 * x
    f_phi = np.rad2deg
    # Where are the lanes?
    lane_width = filter.lanewidth
    d_max = filter.d_max
    d_min = filter.d_min
    phi_max = filter.phi_max
    phi_min = filter.phi_min
    delta_d = filter.delta_d
    delta_phi = filter.delta_phi

    # note transpose
    belief = filter.belief.copy()
    zeros = belief == 0

    belief[zeros] = np.nan

    belief_image = scale(belief, min_value=0)

    x = f_d(filter.d_pcolor)
    y = f_phi(filter.phi_pcolor)

    z = belief_image[:, :, 0]  # just R component
    z = ma.masked_array(z, zeros)

    plt.pcolor(x, y, np.ones(z.shape), cmap="Pastel1")

    plt.pcolor(x, y, z, cmap="gray")

    if other_phi is not None:
        for _phi, _d in zip(other_phi, other_d):
            plt.plot(
                f_d(_d),
                f_phi(_phi),
                "go",
                markersize=15,
                markeredgecolor="none",
                markeredgewidth=3,
                markerfacecolor="blue",
            )

    plt.plot(
        f_d(d),
        f_phi(phi),
        "go",
        markersize=20,
        markeredgecolor="magenta",
        markeredgewidth=3,
        markerfacecolor="none",
    )

    plt.plot(
        f_d(d),
        f_phi(phi),
        "o",
        markersize=2,
        markeredgecolor="none",
        markeredgewidth=0,
        markerfacecolor="magenta",
    )

    W = f_d(lane_width / 2)
    width_white = f_d(filter.linewidth_white)
    width_yellow = f_d(filter.linewidth_yellow)

    plt.plot([-W, -W], [f_phi(phi_min), f_phi(phi_max)], "w-")
    plt.plot([-W - width_white, -W - width_white], [f_phi(phi_min), f_phi(phi_max)], "k-")
    plt.plot([0, 0], [f_phi(phi_min), f_phi(phi_max)], "k-")
    plt.plot([+W, +W], [f_phi(phi_min), f_phi(phi_max)], "-", color="yellow")
    plt.plot([+W + width_yellow, +W + width_yellow], [f_phi(phi_min), f_phi(phi_max)], "-",
             color="yellow")
    s = ""
    s += f"status = {filter.status.value}"
    s += f"\nphi = {f_phi(phi):.1f} deg"
    s += f"\nd = {f_d(d):.1f} cm"
    s += f"\nentropy = {filter.get_entropy():.4f}"
    s += f"\nmax = {belief.max():.4f}"
    s += f"\nmin = {belief.min():.4f}"

    if other_phi is not None:
        s += "\n Other answers:"
        for _phi, _d in zip(other_phi, other_d):
            s += f"\nphi = {f_phi(_phi):.1f} deg"
            s += f"\nd = {f_d(_d):.1f} cm"

    y = f_phi(phi_max) - 10
    args = dict(rotation=-90, color="white")
    annotate = True
    if annotate:
        plt.annotate(s, xy=(0.05, 0.99), xycoords="figure fraction")
        plt.annotate("in middle of right lane", xy=(0, y), **args)
        plt.annotate("on right white tape", xy=(-W, y), **args)
        plt.annotate("on left yellow tape", xy=(+W, y), **args)
        plt.annotate("in other lane", xy=(+W * 1.3, y), **args)

    plt.axis([f_d(d_min), f_d(d_max), f_phi(phi_min), f_phi(phi_max)])

    plt.ylabel(f"phi: orientation (deg); cell = {f_phi(delta_phi):.1f} deg")
    plt.xlabel(f"d: distance from center line (cm); cell = {f_d(delta_d):.1f} cm")

    plt.gca().invert_xaxis()

    bgr = _plt_to_bgr(figure, dpi)
    plt.close()
    return bgr


def _plt_to_bgr(figure, dpi) -> BGRImage:
    """Convert a Matplotlib figure to a PIL Image and return it"""
    buf = io.BytesIO()
    plt.savefig(buf, format="png", dpi=dpi, bbox_inches='tight', pad_inches=0.15,
                transparent=True, facecolor=figure.get_facecolor())
    buf.seek(0)
    png = np.asarray(bytearray(buf.read()), dtype=np.uint8)
    return cv2.imdecode(png, cv2.IMREAD_COLOR)


__all__ = [
    "plot_belief",
    "plot_d_phi"
]



import time
import logging
from threading import Semaphore
from typing import Tuple, Optional

import numpy as np

from dt_state_estimation.wheel_odometry.types import \
    IWheelOdometer, \
    Pose2DEstimate, \
    Velocity2DEstimate
from dt_state_estimation.wheel_odometry.utils import angle_clamp


class WheelOdometer(IWheelOdometer):
    """
    Performs odometry estimation using data from wheel encoders (aka deadreckoning).
    """

    def __init__(self, ticks_per_meter: float, wheel_baseline: float):
        super(WheelOdometer, self).__init__(ticks_per_meter, wheel_baseline)
        # current pose
        self._pose = Pose2DEstimate(0, 0, 0, 0)
        # linear/angular velocity
        self._velocity = Velocity2DEstimate(0, 0, 0)
        # temporary data
        self._left_ticks_last = None
        self._right_ticks_last = None
        self._timestamp_last = None
        self._has_estimate = None
        # others
        self._lock = Semaphore()
        self._logger = logging.getLogger("WheelOdometer")

    def initialize(self):
        pass

    def update(self, left_ticks: int, right_ticks: int, timestamp: float = None):
        if self._timestamp_last is None:
            # move cursor forward
            self._left_ticks_last = left_ticks
            self._right_ticks_last = right_ticks
            self._timestamp_last = timestamp
            return

        # timestamp is NOW if not given
        timestamp = timestamp if timestamp is not None else time.time()

        # compute delta_t between this reading and the previous
        dt = timestamp - self._timestamp_last

        # compute the motion of left and right wheels in number of ticks
        left_delta_ticks = left_ticks - self._left_ticks_last
        right_delta_ticks = right_ticks - self._right_ticks_last

        # compute the motion of left and right wheels in meters traveled
        left_distance = left_delta_ticks / self.ticks_per_meter
        right_distance = right_delta_ticks / self.ticks_per_meter

        # displacement in body-relative x-direction (assuming differential drive)
        delta_x = (left_distance + right_distance) / 2

        # change in heading
        delta_theta = (right_distance - left_distance) / self.wheel_baseline

        if dt < 1e-6:
            self._logger.warning(f"Time between readings ({dt:.5f}) is too small. Ignoring")
            return

        # update internal state
        with self._lock:
            # linear and angular velocities
            self._velocity.v = delta_x / dt
            self._velocity.w = delta_theta / dt

            self._logger.debug(
                f"Time = {timestamp} s; Dt = {dt:.5f} s;\n "
                f"\tLeft wheel: {left_ticks} ticks; {left_distance} meters;\n "
                f"\tRight wheel: {right_ticks} ticks; {right_distance} meters;\n "
                f"\tv: {self._velocity.v:.4f} m/s;\n "
                f"\tw: {np.rad2deg(self._velocity.w)} deg/s;"
            )

            print(delta_x, self._pose.theta, np.cos(self._pose.theta), np.sin(self._pose.theta))

            # update pose
            self._pose.theta = angle_clamp(self._pose.theta + delta_theta)
            self._pose.x = self._pose.x + delta_x * np.cos(self._pose.theta)
            self._pose.y = self._pose.y + delta_x * np.sin(self._pose.theta)

            # move cursor forward
            self._left_ticks_last = left_ticks
            self._right_ticks_last = right_ticks
            self._timestamp_last = timestamp
            self._has_estimate = True

    def get_estimate(self) -> Tuple[Optional[Pose2DEstimate], Optional[Velocity2DEstimate]]:
        if not self._has_estimate:
            return None, None
        with self._lock:
            return self._pose.copy(), self._velocity.copy()



import dataclasses
from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import List, Tuple

import numpy as np


class LaneFilterStatus(Enum):
    UNKNOWN = "UNKNOWN"
    LOST = "LOST"
    GOOD = "GOOD"
    STRUGGLING = "STRUGGLING"


class SegmentColor(Enum):
    WHITE = "white"
    YELLOW = "yellow"
    RED = "red"


@dataclasses.dataclass
class SegmentPoint:
    x: float
    y: float

    def as_array(self) -> np.ndarray:
        return np.array([self.x, self.y])


@dataclasses.dataclass
class Segment:
    points: List[SegmentPoint]
    color: SegmentColor


class ILaneFilter(metaclass=ABCMeta):

    def __init__(self,
                 mean_d_0: float = 0,
                 mean_phi_0: float = 0,
                 sigma_d_0: float = 0.1,
                 sigma_phi_0: float = 0.1,
                 delta_d: float = 0.02,
                 delta_phi: float = np.deg2rad(5),
                 d_max: float = 0.3,
                 d_min: float = -0.15,
                 phi_min: float = -np.deg2rad(85),
                 phi_max: float = np.deg2rad(85),
                 cov_v: float = 0.5,
                 linewidth_white: float = 0.05,
                 linewidth_yellow: float = 0.025,
                 lanewidth: float = 0.23,
                 min_max: float = 0.1,
                 sigma_d_mask: float = 1.0,
                 sigma_phi_mask: float = 2.0,
                 curvature_res: float = 0,
                 range_min: float = 0.2,
                 range_est: float = 0.33,
                 range_max: float = 0.6,
                 curvature_right: float = -0.054,
                 curvature_left: float = 0.025,
                 encoder_resolution: float = 135.0,
                 wheel_baseline: float = 0.1,
                 wheel_radius: float = 0.0318,
                 ):
        # store parameters
        self.mean_d_0: float = mean_d_0
        self.mean_phi_0: float = mean_phi_0
        self.sigma_d_0: float = sigma_d_0
        self.sigma_phi_0: float = sigma_phi_0
        self.delta_d: float = delta_d
        self.delta_phi: float = delta_phi
        self.d_max: float = d_max
        self.d_min: float = d_min
        self.phi_min: float = phi_min
        self.phi_max: float = phi_max
        self.cov_v: float = cov_v
        self.linewidth_white: float = linewidth_white
        self.linewidth_yellow: float = linewidth_yellow
        self.lanewidth: float = lanewidth
        self.min_max: float = min_max
        self.sigma_d_mask: float = sigma_d_mask
        self.sigma_phi_mask: float = sigma_phi_mask
        self.curvature_res: float = curvature_res
        self.range_min: float = range_min
        self.range_est: float = range_est
        self.range_max: float = range_max
        self.curvature_right: float = curvature_right
        self.curvature_left: float = curvature_left
        self.encoder_resolution: float = encoder_resolution
        self.wheel_baseline: float = wheel_baseline
        self.wheel_radius: float = wheel_radius
        # public objects
        self.d, self.phi = \
            np.mgrid[
                self.d_min: self.d_max: self.delta_d,
                self.phi_min: self.phi_max: self.delta_phi
            ]

        self.d_pcolor, self.phi_pcolor = \
            np.mgrid[
                self.d_min: (self.d_max + self.delta_d): self.delta_d,
                self.phi_min: (self.phi_max + self.delta_phi): self.delta_phi,
            ]

        self.belief = np.empty(self.d.shape)

        self.mean_0 = [self.mean_d_0, self.mean_phi_0]
        self.cov_0 = [[self.sigma_d_0, 0], [0, self.sigma_phi_0]]
        self.cov_mask = [self.sigma_d_mask, self.sigma_phi_mask]
        # private objects
        self._status: LaneFilterStatus = LaneFilterStatus.UNKNOWN

    @property
    def status(self) -> LaneFilterStatus:
        return self._status

    @abstractmethod
    def initialize(self):
        pass

    @abstractmethod
    def predict(self, encoder_left_ticks, encoder_right_ticks):
        pass

    @abstractmethod
    def update(self, segment_list):
        """
        segment list: a list of Segment objects
        """

    @abstractmethod
    def get_estimate(self) -> Tuple[float, float]:
        """
        Returns a tuple of two estimated values, `d` (lateral offset) and
        `phi` (heading offset).
        """

    @abstractmethod
    def get_entropy(self) -> float:
        pass

    @abstractmethod
    def get_max(self) -> float:
        pass

    @abstractmethod
    def get_inlier_segments(self, segments: List[Segment], d_max, phi_max):
        pass


__all__ = [
    "LaneFilterStatus",
    "ILaneFilter",
    "Segment",
    "SegmentColor",
    "SegmentPoint"
]



import dataclasses
from abc import ABCMeta, abstractmethod
from typing import Tuple, Optional

import numpy as np

import transformations as tr


@dataclasses.dataclass
class Pose2DEstimate:
    x: float
    y: float
    theta: float
    # time of the estimate
    time: float

    @property
    def q(self) -> np.ndarray:
        return tr.quaternion_from_euler(0, 0, self.theta)

    def copy(self) -> 'Pose2DEstimate':
        return Pose2DEstimate(**dataclasses.asdict(self))

    def __str__(self):
        return f"P(x={self.x}, y={self.y}, theta={self.theta})"


@dataclasses.dataclass
class Velocity2DEstimate:
    v: float
    w: float
    # time of the estimate
    time: float

    def copy(self) -> 'Velocity2DEstimate':
        return Velocity2DEstimate(**dataclasses.asdict(self))

    def __str__(self):
        return f"P(v={self.v}, w={self.w})"


class IWheelOdometer(metaclass=ABCMeta):

    def __init__(self,
                 ticks_per_meter: float,
                 wheel_baseline: float
                 ):
        """
        Performs odometry estimation using data from wheel encoders (aka deadreckoning).

        Args:
            ticks_per_meter (:obj:`int`):   Number of ticks generated by the encoder for
                                            each meter traveled by the wheel;

            wheel_baseline (:obj:`float`):  Distance between the center of the two wheels along
                                            the axle (in meters);

        """
        # store parameters
        self.ticks_per_meter: float = ticks_per_meter
        self.wheel_baseline: float = wheel_baseline

    @abstractmethod
    def initialize(self):
        pass

    @abstractmethod
    def update(self, left_ticks: int, right_ticks: int, timestamp: float = None):
        """
        Args:
            left_ticks `int`:   Number of ticks counted so far by the left wheel encoder.
            right_ticks `int`:  Number of ticks counted so far by the right wheel encoder.
            timestamp `float`:  Time in seconds when the reading happened, default is NOW.

        """

    @abstractmethod
    def get_estimate(self) -> Tuple[Optional[Pose2DEstimate], Optional[Velocity2DEstimate]]:
        """
        Returns a pose estimate and a velocity estimate based on the data processed so far.
        """


__all__ = [
    "IWheelOdometer",
    "Pose2DEstimate",
    "Velocity2DEstimate"
]
